#include "shared.h"
#include <stdlib.h>
#include <stdio.h>

#define ARG_REQUIRED 4 

Features *extractKmers(int **S, int *len, int nStr, int k)
{
  int i, j, jj;
  int n, nfeat;
  int *group;
  int *features;
  int *s;
  int c;
  Features *F;

  nfeat = 0;
  for (i = 0; i < nStr; ++i)
    nfeat += len[i] - k + 1 > 0 ? len[i] - k + 1 : 0;
  group = (int *)malloc(nfeat*sizeof(int));
  features = (int *)malloc(nfeat*k*sizeof(int *));
  c = 0;
  for (i = 0; i < nStr; ++i)
  {
    n = len[i] - k + 1;
    s = S[i];
    for (j = 0; j < n; ++j)
    {
      for (jj = 0; jj < k; ++jj)
        features[c + jj*nfeat] = s[j+jj];
      group[c] = i;
      c++;
    }
  }
  if (nfeat != c)
  {
    printf("Something is wrong...\n");
    exit(1);
  }
  F = (Features *)malloc(sizeof(Features));
  (*F).features = features;
  (*F).group = group;
  (*F).n = nfeat;
  return F;
}

int help()
{
 printf("Usage: wildcardKernel <Sequence-file> <k> <w> <Alphabet-size>\n");
 printf("\t Sequence-file : file with sequence data\n");
 printf("\t k : kernel parameter (k-mer length), k>0\n");
 printf("\t w : kernel parameter (max. number of wildcards), 0<=w<k\n");
 printf("\t Alphabet-size: size of the sequence alphabet, >0\n");
 printf("\t IMPORTANT: sequence elements are assumed to be\n\tin the range [0,AlphabetSize - 1]\n");
  return 1;
}

char isVerbose;
int *wildcardKernel(Features *features, int k, int w, int nStr, int na, int *len);

int main (int argc, char **argv)
{
  int k, w, na;
  char *filename;
  int i;
  int *len;
  int **S, *K;
  int nStr, nfeat;
  int maxIdx;
  char *kernelfilename;
  Features *features;

  isVerbose = 0;

  if (argc != ARG_REQUIRED+1)
  {
    return help();
  }

  filename = argv[1];
  k = atoi( argv[2] );
  w = atoi( argv[3] );
  na = atoi( argv[4] );

  if ( !(k > 0 && w >= 0 && na > 0 && w < k) )
  {
    printf("Error in parameter settings\n");
    return help();
  }
     /* read strings */
     /* use line feeds to end lines */ 
     len = (int *)malloc(MAXNSTR*sizeof(int));
     printf("Reading %s\n",filename);
     S = loadStrings(filename, len, &nStr);
     printf("Read %d strings\n", nStr);
     features = extractKmers(S,len,nStr,k);
     nfeat = (*features).n;
     printf("Extracted %d features\n", nfeat);
     fflush( stdout );
    
     for (i = 0; i < nStr; ++i)
       free(S[i]);
     free(S);
     free(len);
 
     /* check sequence elements range */
     maxIdx = 0;
     for (i = 0; i < nfeat*k; ++i)
       if (maxIdx < (*features).features[i])
         maxIdx = (*features).features[i];
     if (na < maxIdx + 1)
     {
       printf("WARNING: sequence elements are outside of the specified range [0,%d]\n",na);
       printf("\t using [0 %d] instead\n", maxIdx+1);
       na = maxIdx + 1;
     }

     printf("Computing (k=%d,w=%d) wildcard kernel\n", k, w);
     K = wildcardKernel(features, k, w, nStr, na, len);

     /* save kernel matrix in a file */
     kernelfilename = "kernel.txt"; 
     saveKernelInt(K,nStr,nStr,"%d ",kernelfilename);

     return 0;
}

int *wildcardKernel(Features *features, int k, int w, int nStr, int na, int *len)
{
     int i, j, ii, jj, nComb, nfeat, flen;
     int *sortIdx;
     int *comb1;
     int *K;
     Combination **combinations;

     nfeat = (*features).n;
     combinations = (Combination **)malloc((w+1)*sizeof(Combination *));
     for (i = 0; i <= w; ++i)
          combinations[i] = getCombinations(k, k-i);
     K = (int *)malloc(nStr*nStr*sizeof(int));
     for (i = 0; i < nStr*nStr; ++i)
        K[i]=0;

     for (j = 0; j <= w; ++j)
     {
       nComb = (int)nchoosek(k, (k-j));
       printf("w=%d (%d iteration(s))\n", j, nComb);
       for (ii = 0; ii < nComb; ++ii)
       {
         flen = k - j; /* feature length */
         comb1 = (int *)malloc(flen*sizeof(int));
         for (jj = 0; jj < combinations[j][0].k; ++jj)
         {
            comb1[jj] = combinations[j][0].comb[ii+jj*(int)combinations[j][0].num_comb];
            if ( isVerbose ) printf("%d ", comb1[jj]);
         }
         if ( isVerbose ) printf("\n");
         if ( isVerbose ) printf("Sorting..."), fflush( stdout );
         sortIdx = cntsrtna_comb((*features).features,flen,nfeat,na,comb1);
         if ( isVerbose ) printf("Counting..."), fflush( stdout );
         countAndUpdate_comb(K,(*features).features,(*features).group,flen,nfeat,nStr,sortIdx,comb1);
         free(sortIdx);
         free(comb1);
       }
     }
     printf("done.");

     return K;
}

