#include "shared.h"
#include <stdlib.h>
#include <stdio.h>

#define ARG_REQUIRED 5 

Features *extractKmers(int **S, int *len, int nStr, int k)
{
  int i, j, jj;
  int n, nfeat;
  int *group;
  int *features;
  int *s;
  int c;
  Features *F;

  nfeat = 0;
  for (i = 0; i < nStr; ++i)
    nfeat += len[i] - k + 1 > 0 ? len[i] - k + 1 : 0;
  group = (int *)malloc(nfeat*sizeof(int));
  features = (int *)malloc(nfeat*k*sizeof(int *));
  c = 0;
  for (i = 0; i < nStr; ++i)
  {
    n = len[i] - k + 1;
    s = S[i];
    for (j = 0; j < n; ++j)
    {
      for (jj = 0; jj < k; ++jj)
        features[c + jj*nfeat] = s[j+jj];
      group[c] = i;
      c++;
    }
  }
  if (nfeat != c)
  {
    printf("Something is wrong...\n");
    exit(1);
  }
  F = (Features *)malloc(sizeof(Features));
  (*F).features = features;
  (*F).group = group;
  (*F).n = nfeat;
  return F;
}

int help()
{
 printf("Usage: wildcardKernel <Sequence-file> <Query-file> <k> <w> <Alphabet-size>\n");
 printf("\t Sequence-file : file with sequence data\n");
 printf("\t k : kernel parameter (k-mer length), k>0\n");
 printf("\t w : kernel parameter (max. number of wildcards), 0<=w<k\n");
 printf("\t Alphabet-size: size of the sequence alphabet, >0\n");
 printf("\t IMPORTANT: sequence elements are assumed to be\n\tin the range [0,AlphabetSize - 1]\n");
  return 1;
}

char isVerbose;
int *wildcardKernelVector(Features *features, int k, int w, int nStr, int na, int *len, int nq);

int main (int argc, char **argv)
{
  int k, w, na;
  char *databasefilename, *queryfilename;
  int i, j;
  int *len, *lend, *lenq;
  int **S, **Sd, **Sq, *K, *K1;
  int nStr, nfeat, nq;
  int maxIdx;
  char *kernelfilename;
  Features *features;

  isVerbose = 0;

  if (argc != ARG_REQUIRED+1)
  {
    return help();
  }

  databasefilename = argv[1];
  queryfilename = argv[2];
  k = atoi( argv[3] );
  w = atoi( argv[4] );
  na = atoi( argv[5] );

  if ( !(k > 0 && w >= 0 && na > 0 && w < k) )
  {
    printf("Error in parameter settings\n");
    return help();
  }
     /* read strings */
    /* use line feeds to end lines */ 
    lend = (int *)malloc(MAXNSTR*sizeof(int));
    printf("Reading %s\n",databasefilename);
    Sd = loadStrings(databasefilename, lend, &nStr);
    printf("Read %d strings\n", nStr);
    lenq = (int *)malloc(MAXNSTR*sizeof(int));
    printf("Reading %s\n", queryfilename);
    Sq = loadStrings(queryfilename, lenq, &nq);
    printf("Read %d strings\n", nq);
    S = (int **)malloc((nStr+nq)*sizeof(int *));
    len = (int *)malloc((nq + nStr)*sizeof(int));
    for (i = 0; i < nq; ++i)
    {
      S[i] = (int *)malloc(lenq[i]*sizeof(int));
      len[i] = lenq[i];
      for (j = 0; j < lenq[i]; ++j) 
        S[i][j] = Sq[i][j];
    }
    for (i = 0; i < nStr; ++i)
    {
      S[i+nq] = (int *)malloc(lend[i]*sizeof(int));
      len[i+nq] = lend[i];
      for (j = 0; j < lend[i]; ++j) 
        S[i+nq][j] = Sd[i][j];
    }

    features = extractKmers(S,len,nStr+nq,k);
    nfeat = (*features).n;
    printf("Extracted %d features\n", nfeat);
    fflush( stdout );

    for (i = 0; i < nStr+nq; ++i)
       free(S[i]);
    free(S);
    for (i = 0; i < nStr; ++i)
      free(Sd[i]);
    free(Sd); 
    for (i = 0; i < nq; ++i)
      free(Sq[i]);
 
     /* check sequence elements range */
     maxIdx = 0;
     for (i = 0; i < nfeat*k; ++i)
       if (maxIdx < (*features).features[i])
         maxIdx = (*features).features[i];
     if (na < maxIdx + 1)
     {
       printf("WARNING: sequence elements are outside of the specified range [0,%d]\n",na);
       printf("\t using [0 %d] instead\n", maxIdx+1);
       na = maxIdx + 1;
     }

    printf("Computing (k=%d,w=%d) wildcard kernel\n", k, w);

    K = wildcardKernelVector(features, k, w, nStr+nq, na, len, nq);
     /* save kernel matrix in a file */
    K1 = (int *)malloc(nStr*nq*sizeof(int));
    for (i = 0; i < nq; ++i)
      for (j = nq; j < nq+nStr; ++j)
        K1[i+(j-nq)*nq] = K[i+j*(nStr+nq)];
    free(K);
    kernelfilename = "kernel.txt"; 
    saveKernelInt(K1,nq,nStr,"%d ",kernelfilename);
    free(K1);
    return 0;
}

int *wildcardKernelVector(Features *features, int k, int w, int nStr, int na, int *len, int nq)
{
     int i, j, ii, jj, nComb, nfeat, flen;
     int *sortIdx;
     int *comb1;
     int *K;
     Combination **combinations;

     nfeat = (*features).n;
     combinations = (Combination **)malloc((w+1)*sizeof(Combination *));
     for (i = 0; i <= w; ++i)
          combinations[i] = getCombinations(k, k-i);
     K = (int *)malloc(nStr*nStr*sizeof(int));
     for (i = 0; i < nStr*nStr; ++i)
        K[i]=0;

     for (j = 0; j <= w; ++j)
     {
       nComb = (int)nchoosek(k, (k-j));
       printf("w=%d (%d iteration(s))\n", j, nComb);
       for (ii = 0; ii < nComb; ++ii)
       {
         flen = k - j; /* feature length */
         comb1 = (int *)malloc(flen*sizeof(int));
         for (jj = 0; jj < combinations[j][0].k; ++jj)
         {
            comb1[jj] = combinations[j][0].comb[ii+jj*(int)combinations[j][0].num_comb];
            if ( isVerbose ) printf("%d ", comb1[jj]);
         }
         if ( isVerbose ) printf("\n");
         if ( isVerbose ) printf("Sorting..."), fflush( stdout );
         sortIdx = cntsrtna_comb((*features).features,flen,nfeat,na,comb1);
         if ( isVerbose ) printf("Counting..."), fflush( stdout );
         countAndUpdate_comb_vector(K,(*features).features,(*features).group,flen,nfeat,nStr,sortIdx,comb1,nq);
         free(sortIdx);
         free(comb1);
       }
     }
     printf("done.");

     return K;
}

